#include "stdafx.h"
#include "cphidget.h"
#include "cusb.h"
#include "csocket.h"
#include "cthread.h"
#include "cphidgetlist.h"

CPhidgetList *ActiveDevices = 0;
CPhidgetList *AttachedDevices = 0;

/* Protects ActiveDevices and AttachedDevices */
int phidgetLockInitialized = PFALSE;
CThread_mutex_t phidgetLock;

char
translate_bool_to_ascii(char value)
{
	switch (value) {
	case PTRUE:
		return '1';
	case PFALSE:
		return '0';
	}
	return '?';
}

int CCONV phidget_name_to_id(const char *name)
{
	int i;
	for(i=0;i<PHIDGET_DEVICE_TYPE_COUNT;i++)
	{
		if(Phid_DeviceName[i] != NULL)
		{
			if(!strcmp(Phid_DeviceName[i],name))
				return i;
		}
	}
	return -1;
}

/* The is a void pointer that gets malloced only of some platforms */
void CPhdigetFHandle_free(void *arg)
{
	#if defined(_MACOSX) || defined(WINCE)
	return;
	#else
	free(arg);
	return;
	#endif
}

//Two different Phidget Handles that refer to the same Phisical Phidget
int CCONV CPhidget_areEqual(void *arg1, void *arg2)
{
	CPhidgetHandle phid1 = (CPhidgetHandle)arg1;
	CPhidgetHandle phid2 = (CPhidgetHandle)arg2;
	
	if(!phid1||!phid2)
		return EPHIDGET_INVALIDARG;
		
	if((phid1->specificDevice == 0?1:(phid2->serialNumber == phid1->serialNumber)) &&
		(phid1->deviceIDSpec == 0?1:(phid2->deviceIDSpec == phid1->deviceIDSpec)) &&
		(phid2->deviceID == phid1->deviceID))
		return 1;
	return 0;
}

//Two identical Phidget Handles (same pointer)
int CCONV CPhidgetHandle_areEqual(void *arg1, void *arg2)
{
	if(arg1 == arg2) return 1;
	return 0;
}

void CCONV CPhidget_free(void *arg)
{
	CPhidgetHandle phid = (CPhidgetHandle)arg;
	if (!phid)
		return;
	if (phid->CPhidgetFHandle)
		CPhdigetFHandle_free(phid->CPhidgetFHandle);

	CThread_mutex_destroy(&phid->lock);
	CThread_mutex_destroy(&phid->writelock);
	CThread_mutex_destroy(&phid->outputLock);
	CThread_destroy_event(&phid->writeAvailableEvent);
	CThread_destroy_event(&phid->writtenEvent);

	free(phid);
	return;
}

int CCONV CPhidget_create(CPhidgetHandle *phid)
{
	CPhidgetHandle temp_phid;

	if (!phid)
		return EPHIDGET_INVALIDARG;
	
	if(!(temp_phid = malloc(sizeof(CPhidget))))
		return EPHIDGET_NOMEMORY;
	ZEROMEM(temp_phid, sizeof(CPhidget));

	CThread_mutex_init(&temp_phid->lock);
	CThread_mutex_init(&temp_phid->writelock);
	CThread_mutex_init(&temp_phid->outputLock);
	CThread_create_event(&temp_phid->writeAvailableEvent);
	CThread_create_event(&temp_phid->writtenEvent);

	CPhidget_setStatus(temp_phid, PHIDGET_NOTATTACHED);

	*phid = temp_phid;
	 
	 return EPHIDGET_OK;
}

const char *CPhidget_strerror(int error)
{
	if (error < 0 || error >= PHIDGET_ERROR_CODE_COUNT)
		return Phid_UnknownErrorDescription;

	return Phid_ErrorDescriptions[error];
}

int CPhidget_isAttached(CPhidgetHandle phid)
{
	int res;

	CThread_mutex_lock(&phid->lock);
	res = phid->status;
	CThread_mutex_unlock(&phid->lock);
	return res == PHIDGET_ATTACHED;
}

void CPhidget_setStatus(CPhidgetHandle phid, int v)
{
	CThread_mutex_lock(&phid->lock);
	phid->status = v;
	CThread_mutex_unlock(&phid->lock);
}

int CPhidget_read(CPhidgetHandle phid)
{
	int result = EPHIDGET_OK;

	if (!phid)
		return EPHIDGET_INVALIDARG;

	if (CPhidget_isAttached(phid)) {
		result = CUSBReadPacket((CPhidgetHandle)phid,
		    phid->lastReadPacket);
		if (result) return result;
		if (phid->fptrData)
			result= phid->fptrData((CPhidgetHandle)phid,
			    phid->lastReadPacket, phid->inputReportByteLength);
		return result;
	}
	return EPHIDGET_NOTATTACHED;
}

int CPhidget_write(CPhidgetHandle phid)
{
	unsigned char buffer[25] = { 0 };
	unsigned int len;
	int result = EPHIDGET_OK;

	CThread_reset_event(&phid->writeAvailableEvent);

	len = sizeof (buffer);
	if ((result = phid->fptrMakePacket((CPhidgetHandle)phid, buffer, &len))
	  != EPHIDGET_OK)
		goto fail;
	// XXX len ignored
	if ((result = CUSBSendPacket((CPhidgetHandle)phid, buffer)) !=
	  EPHIDGET_OK)
		goto fail;
fail:
	/*
	 * under all circumstances (especially failure), signal
	 * waiting writers to loop around and check device status.
	 */
	CThread_set_event(&phid->writtenEvent);
	return result;
}

//Begin exported functions
int CCONV
CPhidget_open(CPhidgetHandle phid, int serialNumber)
{
	int result = 0;
	if (!phid || serialNumber < -1)
		return EPHIDGET_INVALIDARG;

	if (serialNumber == -1)
		phid->specificDevice = FALSE;
	else
		phid->specificDevice = TRUE;
	phid->serialNumber = serialNumber;

	result = RegisterLocalDevice(phid);

	return result; 
}

int CCONV
CPhidget_close(CPhidgetHandle phid)
{ 
	int result = EPHIDGET_OK;
	struct sockaddr name;
	struct sockaddr_in *name_in;
	socklen_t namelen = sizeof(name);

	char key[1024], val[6];

	if (!phid)
		return EPHIDGET_INVALIDARG;

	if(phid->networkInfo)
	{
		if(CPhidgetSocketClient_isAttached(phid->networkInfo))
		{
			getsockname(phid->networkInfo->socket, &name, &namelen);
			name_in = (struct sockaddr_in *)&name;

			snprintf(key, sizeof(key), "/PCK/Client/%s/%d/%s/%ld", inet_ntoa(name_in->sin_addr), (int)name_in->sin_port, 
				Phid_DeviceName[phid->deviceID], phid->serialNumber);
			snprintf(val, sizeof(val), "Close");
			pdc_async_set(phid->networkInfo->pdcs, key, val, PTRUE, internal_async_network_error_handler, phid);
		}
		CThread_mutex_lock(&serverLock);
		unregisterRemotePhidget(phid);
		CThread_mutex_unlock(&serverLock);
		CPhidget_setStatus(phid, PHIDGET_NOTATTACHED);
		phid->keyCount = 0;
	}
	else
	{
		CList_removeFromList((CListHandle *)&ActiveDevices, phid, CPhidget_areEqual, FALSE, CPhidget_free);
		if (CPhidget_isAttached(phid)) {
			phid->writeStopFlag = PTRUE;
			CThread_join(&phid->writeThread);
			result = CUSBCloseHandle(phid);
			CThread_join(&phid->readThread);
		}
	}
	return result;
}

/*
#ifdef DEBUG
static FILE *file = 0;
void * phidget_malloc(size_t size, const char *source, int line) {
	void *ptr = Malloc(size);
	if(!file)
		file = fopen("memleak.log","w");
	fprintf(file, "%x malloc - %s:%d\n", ptr, source, line); fflush(file);
	return ptr;
}
void phidget_free(void *ptr, const char *source, int line) {
	if(!file)
		file = fopen("memleak.log","w");
	fprintf(file, "%x free - %s:%d\n", ptr, source, line); fflush(file);
	free(ptr);
}
#endif*/

//TODO: maybe this should take (CPhidgetHandle *) so that we can nulify the pointer
// (and handle multiple calls to CPhidget_delete)
int CCONV
CPhidget_delete(CPhidgetHandle phid)
{
	CPhidget_free(phid);
	return EPHIDGET_OK;
}

int CCONV
CPhidget_set_OnDetach_Handler(CPhidgetHandle phid,
    int(CCONV *fptr)(CPhidgetHandle, void *), void *userPtr)
{
	if (!phid)
		return EPHIDGET_INVALIDARG;
	phid->fptrDetach = fptr;
	phid->fptrDetachptr = userPtr;
	return EPHIDGET_OK;
}

int CCONV
CPhidget_set_OnAttach_Handler(CPhidgetHandle phid,
    int(CCONV *fptr)(CPhidgetHandle, void *), void *userPtr)
{
	if (!phid)
		return EPHIDGET_INVALIDARG;
	phid->fptrAttach = fptr;
	phid->fptrAttachptr = userPtr;
	return EPHIDGET_OK;
}

int CCONV
CPhidget_set_OnError_Handler(CPhidgetHandle phid,
    int(CCONV *fptr)(CPhidgetHandle, void *, int, const char *), void *userPtr)
{
	if (!phid)
		return EPHIDGET_INVALIDARG;
	phid->fptrError = fptr;
	phid->fptrErrorptr = userPtr;
	return EPHIDGET_OK;
}

int CCONV
CPhidget_getDeviceName(CPhidgetHandle phid, const char **buffer)
{
	if (!phid)
		return EPHIDGET_INVALIDARG;
	*buffer = (char *)Phid_DeviceSpecificName[phid->deviceIDSpec];
	return EPHIDGET_OK;
}

int CCONV
CPhidget_getSerialNumber(CPhidgetHandle phid, int *serialNumber)
{
	if (!phid)
		return EPHIDGET_INVALIDARG;
	*serialNumber = phid->serialNumber;
	return EPHIDGET_OK;
}

int CCONV
CPhidget_getDeviceVersion(CPhidgetHandle phid, int *devVer)
{
	if (!phid)
		return EPHIDGET_INVALIDARG;
	*devVer = phid->deviceVersion;
	return EPHIDGET_OK;
}

/* for now this just returns the attached bit of the status variable */
int CCONV
CPhidget_getDeviceStatus(CPhidgetHandle phid, int *status)
{
	if (!phid)
		return EPHIDGET_INVALIDARG;
	if (!status)
		return EPHIDGET_INVALIDARG;
	*status = CPhidget_isAttached(phid);
	return EPHIDGET_OK;
}

/* for now this just returns the attached bit of the status variable */
int CCONV
CPhidget_getServerStatus(CPhidgetHandle phid, int *status)
{
	if (!phid)
		return EPHIDGET_INVALIDARG;
	if (!status)
		return EPHIDGET_INVALIDARG;
	if (!phid->networkInfo)
		return EPHIDGET_INVALIDARG;

	CPhidgetSocketClient_isAttached(phid->networkInfo);

	return EPHIDGET_OK;
}


int CCONV
CPhidget_getLibraryVersion(const char **buffer)
{
	if (!buffer)
		return EPHIDGET_INVALIDARG;
	*buffer = LibraryVersion;
	return EPHIDGET_OK;
}

int CCONV
CPhidget_getDeviceType(CPhidgetHandle phid, const char **buffer)
{
	if (!buffer)
		return EPHIDGET_INVALIDARG;
	if (!phid)
		return EPHIDGET_INVALIDARG;
	*buffer = phid->deviceType;
	return EPHIDGET_OK;
}

int CCONV
CPhidget_getDeviceLabel(CPhidgetHandle phid, const char **buffer)
{
	if (!buffer)
		return EPHIDGET_INVALIDARG;
	if (!phid)
		return EPHIDGET_INVALIDARG;
	*buffer = (char *)phid->label;
	return EPHIDGET_OK;
}

int CCONV
CPhidget_setDeviceLabel(CPhidgetHandle phid, const char *buffer)
{
	if(phid->networkInfo)
	{
		char key[1024];
		if(!CPhidgetSocketClient_isAttached(phid->networkInfo))
			return EPHIDGET_NETWORK_NOTCONNECTED;
		snprintf(key, sizeof(key), "/PCK/%s/%ld/Label", phid->deviceType, phid->serialNumber);
		pdc_async_set(phid->networkInfo->pdcs, key, buffer, PFALSE, internal_async_network_error_handler, phid);
		return EPHIDGET_OK;
	}
	else
	{
#ifdef _WINDOWS
	return EPHIDGET_UNSUPPORTED;
#else
	int i;
	char buffer2[22];
	if (!buffer)
		return EPHIDGET_INVALIDARG;
	if (!phid)
		return EPHIDGET_INVALIDARG;
	if (strlen(buffer) <= 10) {
		buffer2[0] = (char)strlen(buffer)*2+2;
		buffer2[1] = 3;
		for (i = 0; i < (int)strlen(buffer); i++) {
			buffer2[2+i*2] = buffer[i];
			buffer2[2+i*2+1]=0;
		}
		if (!CUSBSetLabel(phid, buffer2)) {
			strcpy(phid->label, buffer);
			return EPHIDGET_OK;
		}
		return EPHIDGET_UNEXPECTED;
	}
	return EPHIDGET_INVALID;
#endif
	}
}

int CCONV
CPhidget_getServerID(CPhidgetHandle phid, const char **buffer)
{
	return EPHIDGET_UNSUPPORTED; //TODO: remove this when ServerID is supported.
	if (!phid || !phid->networkInfo)
		return EPHIDGET_INVALIDARG;
	if(!phid->networkInfo->serverID)
		return EPHIDGET_NOTATTACHED;
	*buffer = (char *)phid->networkInfo->serverID;
	return EPHIDGET_OK;
}

int CCONV
CPhidget_getServerAddress(CPhidgetHandle phid, const char **ipAddr, int *port)
{
	if (!phid || !phid->networkInfo)
		return EPHIDGET_INVALIDARG;
	if(!phid->networkInfo->address || !phid->networkInfo->port)
		return EPHIDGET_NOTATTACHED;
	*ipAddr = (char *)phid->networkInfo->address;
	*port = strtol(phid->networkInfo->port, NULL, 10);
	return EPHIDGET_OK;
}

int CCONV CPhidget_getErrorDescription(int ErrorCode, const char **buf)
{
	TESTPTR(buf)
	*buf = CPhidget_strerror(ErrorCode);
	return EPHIDGET_OK;
}

int CCONV CPhidget_calibrate_gainoffset(CPhidgetHandle phid, int Index, unsigned short offset, unsigned long gain)
{
	int j, result;
	unsigned char buffer[8];
	if (!phid)
		return EPHIDGET_INVALIDARG;
	for (j = 0; j<8; j++) buffer[j] = 0;
	
	buffer[0] = 0x72;
	
	buffer[1] = 0 + Index;
	buffer[2] = (unsigned char)(offset >> 8);
	if ((result = CUSBSendPacket((CPhidgetHandle)phid, buffer)) != EPHIDGET_OK)
		return result;

	buffer[1] = 1 + Index;
	buffer[2] = (unsigned char)(offset & 0xff);
	if ((result = CUSBSendPacket((CPhidgetHandle)phid, buffer)) != EPHIDGET_OK)
		return result;

	buffer[1] = 2 + Index;
	buffer[2] = (unsigned char)(gain >> 16);
	if ((result = CUSBSendPacket((CPhidgetHandle)phid, buffer)) != EPHIDGET_OK)
		return result;

	buffer[1] = 3 + Index;
	buffer[2] = (unsigned char)(gain >> 8);
	if ((result = CUSBSendPacket((CPhidgetHandle)phid, buffer)) != EPHIDGET_OK)
		return result;

	buffer[1] = 4 + Index;
	buffer[2] = (unsigned char)(gain);
	if ((result = CUSBSendPacket((CPhidgetHandle)phid, buffer)) != EPHIDGET_OK)
		return result;

	return EPHIDGET_OK;
}

//0 ms = infinite timeout - note that it only checks the attach variable every 10ms
int CCONV CPhidget_waitForAttachment(CPhidgetHandle phid, int milliseconds)
{
	long duration = 0;
	TIME	start;

#ifdef _WINDOWS
	TIME now;
	FILETIME nowft, oldft, resft;
	ULARGE_INTEGER nowul, oldul, resul;
#else
	struct timeval now;
#endif

	if (!phid)
		return EPHIDGET_INVALIDARG;
	if(milliseconds)
	{
#ifdef _WINDOWS
		GetSystemTime(&start);
#else
		gettimeofday(&start,NULL);
#endif
	}

	while(!CPhidget_isAttached(phid))
	{
		if(milliseconds)
		{
#ifdef _WINDOWS
			GetSystemTime(&now);
			SystemTimeToFileTime(&now, &nowft);
			SystemTimeToFileTime(&start, &oldft);

			nowul.HighPart = nowft.dwHighDateTime;
			nowul.LowPart = nowft.dwLowDateTime;
			oldul.HighPart = oldft.dwHighDateTime;
			oldul.LowPart = oldft.dwLowDateTime;

			resul.HighPart = nowul.HighPart - oldul.HighPart;
			resul.LowPart = nowul.LowPart - oldul.LowPart;

			resft.dwHighDateTime = resul.HighPart;
			resft.dwLowDateTime = resul.LowPart;

			duration = (resft.dwLowDateTime/10000);
#else
			gettimeofday(&now, NULL);
			duration = (now.tv_sec - start.tv_sec)*1000 + ((now.tv_usec-start.tv_usec)/1000);
#endif
			if(duration > milliseconds)
				return EPHIDGET_TIMEOUT;
		}
		SLEEP(10);
	}
	return EPHIDGET_OK;
}

void throw_error_event(CPhidgetHandle phid, const char *error, int errcode)
{
	if(phid && phid->fptrError)
	{
		phid->fptrError(phid, phid->fptrErrorptr, errcode, error);
		return;
	}
	WARN("Got an async error, aborting: %d: %s\n\tTip: Set up an error handler to catch this without an abort.", errcode, error);
	abort();
}
