#include "stdafx.h"
#include "csocket.h"
#include "csocketevents.h"
#include "cphidgetlist.h"
#include "cphidgetmanager.h"
#include "cphidgetdictionary.h"

typedef struct _CServerInfo
{
	CPhidgetSocketClientHandle server;
	CPhidgetListHandle phidgets;
	CPhidgetManagerListHandle managers;
	CPhidgetDictionaryListHandle dictionaries;
} CServerInfo, *CServerInfoHandle;

typedef struct _CServerList
{
	struct _CServerList *next;
	CServerInfoHandle serverInfo;
} CServerList, *CServerListHandle;

CServerList *servers = NULL;
/* Protects servers */
int serverLockInitialized = PFALSE;
CThread_mutex_t serverLock;

int NetworkInitialized = PFALSE;

#ifdef _WINDOWS
int start_WSA_server()
{
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;

	wVersionRequested = MAKEWORD( 2, 2 );

	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 ) {
		DPRINT("Cannot start WSA\n");
		return EPHIDGET_NETWORK;
	}
	return EPHIDGET_OK;
}
#endif

int CServerInfo_areEqual(void *arg1, void *arg2)
{
	CServerInfoHandle server1 = (CServerInfoHandle)arg1;
	CServerInfoHandle server2 = (CServerInfoHandle)arg2;
	
	if(!server1 || !server2 || !server1->server || !server2->server)
		return EPHIDGET_INVALIDARG;
	
	return CPhidgetSocketClient_areEqual(server1->server, server2->server);
}

void CServerInfo_free(void *arg)
{
	CServerInfoHandle server = (CServerInfoHandle)arg;
	
	if(!server)
		return;
	if(server->server)
		CPhidgetSocketClient_free(server->server);
	//CList_emptyList((CListHandle *)&server->listen_ids, PFALSE, NULL);
	free(server);
	return;
}

int CCONV CPhidgetSocketClient_create(CPhidgetSocketClientHandle *arg)
{
	CPhidgetSocketClientHandle socket_client;
	
	if(!(socket_client = malloc(sizeof(CPhidgetSocketClient))))
		return EPHIDGET_NOMEMORY;
	ZEROMEM(socket_client, sizeof(CPhidgetSocketClient));

	CThread_mutex_init(&socket_client->lock);
	CThread_mutex_init(&socket_client->pdc_lock);

	CPhidgetSocketClient_setStatus(socket_client, PHIDGET_NOTATTACHED);

	*arg = socket_client;
	 
	 return EPHIDGET_OK;
}

void CCONV CPhidgetSocketClient_free(void *arg)
{
	CPhidgetSocketClientHandle socket_client = (CPhidgetSocketClientHandle)arg;

	if(!socket_client) return;
	if(socket_client->port) free(socket_client->port);
	if(socket_client->address) free(socket_client->address);
	if(socket_client->serverID) free(socket_client->serverID);
	if(socket_client->pdcs) free(socket_client->pdcs);

	CThread_mutex_destroy(&socket_client->lock);
	CThread_mutex_destroy(&socket_client->pdc_lock);

	free(socket_client);
	 
	return;
}

int CCONV CPhidgetSocketClient_areEqual(void *arg1, void *arg2)
{
	CPhidgetSocketClientHandle socket1 = (CPhidgetSocketClientHandle)arg1;
	CPhidgetSocketClientHandle socket2 = (CPhidgetSocketClientHandle)arg2;
	
	if(!socket1 || !socket2)
		return EPHIDGET_INVALIDARG;
	if(!socket1->address || !socket2->address || !socket1->port || !socket2->port)
		return EPHIDGET_INVALIDARG;
	
	if(!(strcmp(socket1->address,socket2->address))
		&& !(strcmp(socket1->port, socket2->port)))
		return 1;
		
	return 0;
}

int CPhidgetSocketClient_isAttached(CPhidgetSocketClientHandle socket)
{
	int res;

	CThread_mutex_lock(&socket->lock);
	res = socket->status;
	CThread_mutex_unlock(&socket->lock);
	return res == PHIDGET_ATTACHED;
}

void CPhidgetSocketClient_setStatus(CPhidgetSocketClientHandle socket, int v)
{
	CThread_mutex_lock(&socket->lock);
	socket->status = v;
	CThread_mutex_unlock(&socket->lock);
}

void internal_async_network_error_handler(const char *error, void *ptr)
{
	CPhidgetHandle phid = (CPhidgetHandle)ptr;
	if(phid && phid->fptrError)
	{
		phid->fptrError(phid, phid->fptrErrorptr, EPHIDGET_NETWORK, error);
		return;
	}
	WARN("Got an async network error: %s\n\tTip: Set up an error handler to catch this properly.", error);
	//abort();
}

int InitializeNetworking()
{
	int res;
	const char *setpattern = "^/PSK/([a-zA-Z_0-9]*)/([0-9]*)/([a-zA-Z_0-9]*)/?([a-zA-Z_0-9]*)/?([a-zA-Z_0-9]*)$";
	const char *managerpattern = "^/PSK/List/([a-zA-Z_0-9]*)/([0-9]*)/([0-9]*)/([0-9]*)/\"(.*)\"$";

#ifdef _WINDOWS
	if (start_WSA_server())
	{
		WARN("Cannot start Windows Sockets");
		return EPHIDGET_NETWORK;
	}
#endif
	
	if (!pdc_init()) {
		return EPHIDGET_UNEXPECTED;
	}

	if ((res = regcomp(&phidgetsetex, setpattern, REG_EXTENDED)) != 0) {
		DPRINT("set command pattern compilation error %d\n",
			res);
		abort();
	}
	if ((res = regcomp(&managerex, managerpattern, REG_EXTENDED)) != 0) {
		DPRINT("set command pattern compilation error %d\n",
			res);
		abort();
	}

	NetworkInitialized = PTRUE;

	return EPHIDGET_OK;
}

void cleanup_after_socket(void *ptr)
{
	CPhidgetSocketClientHandle serverInfo = ptr;
	CServerList *travServers;
	CPhidgetListHandle travPhidgets;
	CPhidgetDictionaryListHandle travDicts;
	CPhidgetManagerListHandle travManagers;

	CPhidgetListHandle detachEvents = NULL;
	CPhidgetListHandle errorEvents = NULL;

	/* For each Phidget associated with this server, send a detach event, and an async error */
	// Note that we leave the unconnected socket in the server list until all of it's references have called close
	CThread_mutex_lock(&serverLock);
	for(travServers = servers; travServers; travServers = travServers->next)
	{
		if(travServers->serverInfo->server->socket == serverInfo->socket)
		{
			CPhidgetSocketClient_setStatus(travServers->serverInfo->server, PHIDGET_NOTATTACHED);
			for(travPhidgets = travServers->serverInfo->phidgets; travPhidgets; travPhidgets = travPhidgets->next)
			{
				CPhidget_setStatus(travPhidgets->phid, PHIDGET_NOTATTACHED);
				if(travPhidgets->phid->fptrDetach)
					CList_addToList((CListHandle *)&detachEvents, travPhidgets->phid, CPhidget_areEqual);
					//travPhidgets->phid->fptrDetach((CPhidgetHandle)travPhidgets->phid, travPhidgets->phid->fptrDetachptr);
				CList_addToList((CListHandle *)&errorEvents, travPhidgets->phid, CPhidget_areEqual);
				//internal_async_network_error_handler("The Network Connection has been Closed", travPhidgets->phid);	
			}
			for(travDicts = travServers->serverInfo->dictionaries; travDicts; travDicts = travDicts->next)
			{
				CPhidget_setStatus((CPhidgetHandle)travDicts->dict, PHIDGET_NOTATTACHED);
				CList_addToList((CListHandle *)&errorEvents, travDicts->dict, CPhidgetDictionary_areEqual);
				//internal_async_network_error_handler("The Network Connection has been Closed", travDicts->dict);
			}
			for(travManagers = travServers->serverInfo->managers; travManagers; travManagers = travManagers->next)
			{
				CPhidget_setStatus((CPhidgetHandle)travManagers->phidm, PHIDGET_NOTATTACHED);
				CList_addToList((CListHandle *)&errorEvents, travManagers->phidm, CPhidgetManager_areEqual);
				//internal_async_network_error_handler("The Network Connection has been Closed", travManagers->phidm);
			}
			break;
		}
	}
	CThread_mutex_unlock(&serverLock);

	for(travPhidgets = detachEvents; travPhidgets; travPhidgets = travPhidgets->next)
	{
		travPhidgets->phid->fptrDetach((CPhidgetHandle)travPhidgets->phid, travPhidgets->phid->fptrDetachptr);
	}
	for(travPhidgets = errorEvents; travPhidgets; travPhidgets = travPhidgets->next)
	{
		internal_async_network_error_handler("The Network Connection has been Closed", travPhidgets->phid);
	}

	CList_emptyList((CListHandle *)&detachEvents, FALSE, NULL);
	CList_emptyList((CListHandle *)&errorEvents, FALSE, NULL);
}

int connectToServer(CPhidgetSocketClientHandle *serverInfo, CServerInfoHandle *foundServer, char *password)
{
	char errdesc[1024];
	int result = EPHIDGET_OK;
	
	CServerInfoHandle newServerInfo;
	
	//Initialize the network if not already done
	if(!NetworkInitialized)
		if((result = InitializeNetworking()))
			return result;
	
	if(!(newServerInfo = malloc(sizeof(CServerInfo))))
		return EPHIDGET_NOMEMORY;
	ZEROMEM(newServerInfo, sizeof(CServerInfo));
		
	newServerInfo->server = *serverInfo;
	
	//check to see if there is already a connection to this server
	result = CList_findInList((CListHandle)servers, newServerInfo, CServerInfo_areEqual, (void **)foundServer);
	switch(result)
	{
		case EPHIDGET_OK: //Found
			CPhidgetSocketClient_free(*serverInfo);
			*serverInfo = (*foundServer)->server;
			free(newServerInfo);
			free(password);
			break;
		case EPHIDGET_NOTFOUND: //Not Found
			if (!stream_server_connect((*serverInfo)->address, (*serverInfo)->port, &(*serverInfo)->socket, errdesc,
				sizeof (errdesc))) {
				DPRINT("connect(%s:%s): %s\n", (*serverInfo)->address, (*serverInfo)->port,
					errdesc);
				CPhidgetSocketClient_setStatus((*serverInfo), PHIDGET_CONNECTIONERROR);
				CList_removeFromList((CListHandle *)&servers, newServerInfo, CServerInfo_areEqual, PTRUE, CServerInfo_free);
				*serverInfo = NULL;
				return EPHIDGET_NETWORK;
			}
			
			if (!((*serverInfo)->pdcs = pdc_session_alloc((*serverInfo)->socket, pu_read, (*serverInfo)->socket, pu_write, pu_close, *serverInfo, cleanup_after_socket, password))) {
				fflush(stderr);
				CPhidgetSocketClient_setStatus((*serverInfo), PHIDGET_CONNECTIONERROR);
				CList_removeFromList((CListHandle *)&servers, newServerInfo, CServerInfo_areEqual, PTRUE, CServerInfo_free);
				*serverInfo = NULL;
				return EPHIDGET_UNEXPECTED;
			}
			
			while(!pdc_get_status((*serverInfo)->pdcs))
			{
				SLEEP(10);
			}

			if(pdc_get_status((*serverInfo)->pdcs) == 2)
			{
				DPRINT("bad password");
				CList_removeFromList((CListHandle *)&servers, newServerInfo, CServerInfo_areEqual, PTRUE, CServerInfo_free);
				*serverInfo = NULL;
				return EPHIDGET_BADPASSWORD;
			}
	
			/* maybe give to user a way to change this? */
			if(!pdc_enable_periodic_reports((*serverInfo)->pdcs, 10, errdesc, sizeof (errdesc)))
			{
				DPRINT("pdc_enable_periodic_reports: %s\n", errdesc);
				CList_removeFromList((CListHandle *)&servers, newServerInfo, CServerInfo_areEqual, PTRUE, CServerInfo_free);
				*serverInfo = NULL;
				return EPHIDGET_UNEXPECTED;
			}
			
			if((result = CList_addToList((CListHandle *)&servers, newServerInfo, CServerInfo_areEqual)))
				return result;
			if((result = CList_findInList((CListHandle)servers, newServerInfo, CServerInfo_areEqual, (void **)foundServer)))
				return result;

			CPhidgetSocketClient_setStatus((*serverInfo), PHIDGET_ATTACHED);
			break;
		default:
			return result;
	}
	return EPHIDGET_OK;
}

int setupKeysAndListeners_phidget(CPhidgetHandle phid, pdc_listen_id_t *id)
{
	char errdesc[1024];
	char listenKey[1024];
	struct sockaddr name;
	struct sockaddr_in *name_in;
	socklen_t namelen = sizeof(name);

	char key[1024], val[1024];

	if(!phid || !phid->networkInfo)
		return EPHIDGET_INVALIDARG;

	//listen for everything to do with this specific phidget
	if(phid->specificDevice)
	{
		snprintf(listenKey, sizeof(listenKey), "^/PSK/%s/%d/", Phid_DeviceName[phid->deviceID], phid->serialNumber);
	}
	else
	{
		snprintf(listenKey, sizeof(listenKey), "^/PSK/%s/", Phid_DeviceName[phid->deviceID]);
	}
	
	CThread_mutex_lock(&phid->networkInfo->pdc_lock);
	if (!(*id = pdc_listen(phid->networkInfo->pdcs, listenKey, network_phidget_event_handler, phid, errdesc, sizeof (errdesc))))
	{
		DPRINT("pdc_listen: %s\n", errdesc);
		CThread_mutex_unlock(&phid->networkInfo->pdc_lock);
		return EPHIDGET_UNEXPECTED;
	}
	CThread_mutex_unlock(&phid->networkInfo->pdc_lock);
	//Open the remote device
	//get socket info
	getsockname(phid->networkInfo->socket, &name, &namelen);
	name_in = (struct sockaddr_in *)&name;
	if(phid->specificDevice)
	{
		snprintf(key, sizeof(key), "/PCK/Client/%s/%d/%s/%d", inet_ntoa(name_in->sin_addr), (int)name_in->sin_port, 
			Phid_DeviceName[phid->deviceID], phid->serialNumber);
	}
	else
	{
		snprintf(key, sizeof(key), "/PCK/Client/%s/%d/%s", inet_ntoa(name_in->sin_addr), (int)name_in->sin_port, 
			Phid_DeviceName[phid->deviceID]);
	}
	snprintf(val, sizeof(val), "Open");
	pdc_async_set(phid->networkInfo->pdcs, key, val, PTRUE, internal_async_network_error_handler, phid);

	return EPHIDGET_OK;
}

int setupKeysAndListeners_manager(CPhidgetManagerHandle phidm, pdc_listen_id_t *id)
{
	char errdesc[1024];
	char listenKey[1024];

	if(!phidm || !phidm->networkInfo)
		return EPHIDGET_INVALIDARG;

	//listen for everything to do with the PhidgetManager
	snprintf(listenKey, sizeof(listenKey), "^/PSK/List/");

	CThread_mutex_lock(&phidm->networkInfo->pdc_lock);
	if (!(*id = pdc_listen(phidm->networkInfo->pdcs, listenKey, network_manager_event_handler, phidm, errdesc, sizeof (errdesc))))
	{
		DPRINT("pdc_listen: %s\n", errdesc);
		CThread_mutex_unlock(&phidm->networkInfo->pdc_lock);
		return EPHIDGET_UNEXPECTED;
	}
	CThread_mutex_unlock(&phidm->networkInfo->pdc_lock);

	return EPHIDGET_OK;
}

int RegisterRemotePhidget(CPhidgetHandle phid)
{
	int result = EPHIDGET_OK, result2;
	CServerInfoHandle foundServer;
	
	result = connectToServer(&phid->networkInfo, &foundServer, phid->password);
	switch(result)
	{
		case EPHIDGET_OK: /* Only need to register this once */
			if((result2 = setupKeysAndListeners_phidget(phid, &phid->listen_id)))
				return result2;	
			if((result = CList_addToList((CListHandle *)&foundServer->phidgets, phid, CPhidget_areEqual)))
				return result;
			break;
		default:
			return result;
	}

	return EPHIDGET_OK;
}

int RegisterRemoteManager(CPhidgetManagerHandle phidm)
{
	int result = EPHIDGET_OK, result2;
	CServerInfoHandle foundServer;
	
	result = connectToServer(&phidm->networkInfo, &foundServer, phidm->password);
	switch(result)
	{
		case EPHIDGET_OK: /* Only need to register this once */
			if((result2 = setupKeysAndListeners_manager(phidm, &phidm->listen_id)))
				return result2;
			if((result = CList_addToList((CListHandle *)&foundServer->managers, phidm, CPhidgetManager_areEqual)))
				return result;
			break;
		default:
			return result;
	}

	return EPHIDGET_OK;
}

int RegisterRemoteDictionary(CPhidgetDictionaryHandle dict)
{
	int result = EPHIDGET_OK;
	CServerInfoHandle foundServer;
	
	result = connectToServer(&dict->networkInfo, &foundServer, dict->password);
	switch(result)
	{
		case EPHIDGET_OK:
			if((result = CList_addToList((CListHandle *)&foundServer->dictionaries, dict, CPhidgetDictionary_areEqual)))
				return result;
			break;
		default:
			return result;
	}

	return EPHIDGET_OK;
}

int unregisterRemotePhidget(CPhidgetHandle phid)
{
	int result = EPHIDGET_OK;
	CServerInfoHandle foundServer;
	char errdesc[1024];
	CServerInfo newServerInfo;
		
	newServerInfo.server = phid->networkInfo;

	//check to see if there is already a connection to this server
	result = CList_findInList((CListHandle)servers, &newServerInfo, CServerInfo_areEqual, (void **)&foundServer);
	switch(result)
	{
		case EPHIDGET_OK: //Found

			if((result = CList_removeFromList((CListHandle *)&foundServer->phidgets, phid, CPhidget_areEqual, PFALSE, NULL)))
				return result;
			
			//no more phidgets references to this server
			if(!foundServer->phidgets)
			{
				//stop reports
				CThread_mutex_lock(&phid->networkInfo->pdc_lock);
				if(!pdc_ignore(foundServer->server->pdcs,phid->listen_id,errdesc,sizeof(errdesc)))
				{
					DPRINT("pdc_ignore: %s\n", errdesc);
				}
				CThread_mutex_unlock(&phid->networkInfo->pdc_lock);
				if(!foundServer->dictionaries && !foundServer->managers)
				{
					CThread_mutex_lock(&phid->networkInfo->pdc_lock);
					if(!(pdc_quit(foundServer->server->pdcs,errdesc,sizeof(errdesc))))
					{
						DPRINT("pdc_quit: %s\n", errdesc);
					}
					CThread_mutex_unlock(&phid->networkInfo->pdc_lock);
					if((result = CList_removeFromList((CListHandle *)&servers, foundServer, CServerInfo_areEqual, PTRUE, CServerInfo_free)))
						return result;
				}
			}

			phid->networkInfo = 0;
			
			break;
		case EPHIDGET_NOTFOUND: //Not Found - That's ok, just means it's already closed
			return EPHIDGET_OK;
		default:
			return result;
	}

	return EPHIDGET_OK;
}

int unregisterRemoteManager(CPhidgetManagerHandle phidm)
{
	int result = EPHIDGET_OK;
	CServerInfoHandle foundServer;
	char errdesc[1024];
	CServerInfo newServerInfo;
		
	newServerInfo.server = phidm->networkInfo;

	//check to see if there is already a connection to this server
	result = CList_findInList((CListHandle)servers, &newServerInfo, CServerInfo_areEqual, (void **)&foundServer);
	switch(result)
	{
		case EPHIDGET_OK: //Found

			if((result = CList_removeFromList((CListHandle *)&foundServer->managers, phidm, CPhidgetManager_areEqual, PFALSE, NULL)))
				return result;
			
			if(!foundServer->managers)
			{
				//stop reports
				CThread_mutex_lock(&phidm->networkInfo->pdc_lock);
				if(!pdc_ignore(foundServer->server->pdcs,phidm->listen_id,errdesc,sizeof(errdesc)))
				{
					DPRINT("pdc_ignore: %s\n", errdesc);
				}
				CThread_mutex_unlock(&phidm->networkInfo->pdc_lock);
				if(!foundServer->dictionaries && !foundServer->phidgets)
				{
					CThread_mutex_lock(&phidm->networkInfo->pdc_lock);
					if(!(pdc_quit(foundServer->server->pdcs,errdesc,sizeof(errdesc))))
					{
						DPRINT("pdc_quit: %s\n", errdesc);
					}
					CThread_mutex_unlock(&phidm->networkInfo->pdc_lock);
					if((result = CList_removeFromList((CListHandle *)&servers, foundServer, CServerInfo_areEqual, PTRUE, CServerInfo_free)))
						return result;
				}
			}

			phidm->networkInfo = 0;
			
			break;
		case EPHIDGET_NOTFOUND: //Not Found - That's ok, just means it's already closed
			return EPHIDGET_OK;
		default:
			return result;
	}

	return EPHIDGET_OK;
}

int unregisterRemoteDictionary(CPhidgetDictionaryHandle dict)
{
	int result = EPHIDGET_OK;
	CServerInfoHandle foundServer;
	char errdesc[1024];
	CServerInfo newServerInfo;
		
	newServerInfo.server = dict->networkInfo;

	//check to see if there is already a connection to this server
	result = CList_findInList((CListHandle)servers, &newServerInfo, CServerInfo_areEqual, (void **)&foundServer);
	switch(result)
	{
		case EPHIDGET_OK: //Found

			if((result = CList_removeFromList((CListHandle *)&foundServer->dictionaries, dict, CPhidgetDictionary_areEqual, PFALSE, NULL)))
				return result;
			
			if(!foundServer->dictionaries)
			{
				CPhidgetDictionaryListenerListHandle trav;
				//stop reports
				for(trav = dict->listeners; trav; trav = trav->next)
				{
					CThread_mutex_lock(&dict->networkInfo->pdc_lock);
					if(!pdc_ignore(foundServer->server->pdcs,trav->listener->listen_id,errdesc,sizeof(errdesc)))
					{
						DPRINT("pdc_ignore: %s\n", errdesc);
					}
					CThread_mutex_unlock(&dict->networkInfo->pdc_lock);
				}
				
				if(!foundServer->phidgets && !foundServer->managers)
				{
					CThread_mutex_lock(&dict->networkInfo->pdc_lock);
					if(!(pdc_quit(foundServer->server->pdcs,errdesc,sizeof(errdesc))))
					{
						DPRINT("pdc_quit: %s\n", errdesc);
					}
					CThread_mutex_unlock(&dict->networkInfo->pdc_lock);
					if((result = CList_removeFromList((CListHandle *)&servers, foundServer, CServerInfo_areEqual, PTRUE, CServerInfo_free)))
						return result;
				}
			}

			dict->networkInfo = 0;
			
			break;
		case EPHIDGET_NOTFOUND: //Not Found - That's ok, just means it's already closed
			return EPHIDGET_OK;
		default:
			return result;
	}

	return EPHIDGET_OK;
}

int CCONV
CPhidget_openRemoteIP(CPhidgetHandle phid, int serialNumber, const char *address,
    int port, const char *password)
{
	int result = EPHIDGET_OK;
	char portString[6];

	if (!phid || serialNumber < -1)
		return EPHIDGET_INVALIDARG;
	if(password) {
		if (strlen(password) > 255)
			return EPHIDGET_INVALIDARG;
		if (!(phid->password = strdup(password)))
			return EPHIDGET_NOMEMORY;
	}

	if (serialNumber == -1)
		phid->specificDevice = FALSE;
	else
		phid->specificDevice = TRUE;
	phid->serialNumber = serialNumber;

	if((result = CPhidgetSocketClient_create(&phid->networkInfo)))
		return result;
	snprintf(portString, sizeof(portString), "%d", port);
	if(!(phid->networkInfo->port = strdup(portString)))
		return EPHIDGET_NOMEMORY;
	if(!(phid->networkInfo->address = strdup(address)))
		return EPHIDGET_NOMEMORY;
	
	//Initialize the server mutex if not already done
	if(!serverLockInitialized)
		CThread_mutex_init(&serverLock);

	CThread_mutex_lock(&serverLock);
	result = RegisterRemotePhidget(phid);
	CThread_mutex_unlock(&serverLock);

	return result;
}

/* Not yet implemented */
int CCONV CPhidget_openRemote(CPhidgetHandle phid, int serial, const char *serverID, const char *passWord)
{
	return EPHIDGET_UNSUPPORTED;
}

int CCONV CPhidgetManager_openRemoteIP(CPhidgetManagerHandle phidm, const char *address,
    int port, const char *password)
{
	int result = EPHIDGET_OK;
	char portString[6];

	if (!phidm)
		return EPHIDGET_INVALIDARG;
	if(password) {
		if (strlen(password) > 255)
			return EPHIDGET_INVALIDARG;
		if (!(phidm->password = strdup(password)))
			return EPHIDGET_NOMEMORY;
	}

	if((result = CPhidgetSocketClient_create(&phidm->networkInfo)))
		return result;
	snprintf(portString, sizeof(portString), "%d", port);
	if(!(phidm->networkInfo->port = strdup(portString)))
		return EPHIDGET_NOMEMORY;
	if(!(phidm->networkInfo->address = strdup(address)))
		return EPHIDGET_NOMEMORY;
	
	phidm->state = PHIDGETMANAGER_ACTIVE;
	//Initialize the server mutex if not already done
	if(!serverLockInitialized)
		CThread_mutex_init(&serverLock);

	CThread_mutex_lock(&serverLock);
	result = RegisterRemoteManager(phidm);
	CThread_mutex_unlock(&serverLock);

	return result;
}

int CCONV CPhidgetManager_openRemote(CPhidgetManagerHandle phidm, const char *serverID, const char *passWord)
{
	return EPHIDGET_UNSUPPORTED;
}

int CCONV CPhidgetDictionary_openRemoteIP(CPhidgetDictionaryHandle dict, const char *address,
    int port, const char *password)
{
	int result = EPHIDGET_OK;
	char portString[6];
	if (!dict)
		return EPHIDGET_INVALIDARG;
	if(password) {
		if (strlen(password) > 255)
			return EPHIDGET_INVALIDARG;
		if (!(dict->password = strdup(password)))
			return EPHIDGET_NOMEMORY;
	}

	if((result = CPhidgetSocketClient_create(&dict->networkInfo)))
		return result;
	snprintf(portString, sizeof(portString), "%d", port);
	if(!(dict->networkInfo->port = strdup(portString)))
		return EPHIDGET_NOMEMORY;
	if(!(dict->networkInfo->address = strdup(address)))
		return EPHIDGET_NOMEMORY;
	
	//Initialize the server mutex if not already done
	if(!serverLockInitialized)
		CThread_mutex_init(&serverLock);

	CThread_mutex_lock(&serverLock);
	result = RegisterRemoteDictionary(dict);
	CThread_mutex_unlock(&serverLock);

	return result;
}

int CCONV CPhidgetDictionary_openRemote(CPhidgetDictionaryHandle dict, const char *serverID, const char *passWord)
{
	return EPHIDGET_UNSUPPORTED;
}
